/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import static java.util.Collections.reverse;
import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createPoint;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getNameOfFirstRelatedLibrary;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.isCorrectReuseElement;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.dnd.ElementDropContext;
import org.fortiss.tooling.base.layout.IAutoLayouter;
import org.fortiss.tooling.base.layout.KielerAutoLayouter;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.kernel.extension.data.IElementCompositionContext;
import org.fortiss.tooling.kernel.extension.data.Prototype;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.service.IElementCompositorService;
import org.fortiss.tooling.kernel.service.IPrototypeService;
import org.fortiss.tooling.kernel.ui.extension.IModelEditor;
import org.fortiss.tooling.kernel.ui.service.IModelEditorBindingService;

import javafx.scene.control.Menu;
import javafx.scene.control.MenuItem;

/**
 * Helper class for creating context menus.
 * 
 * @author hoelzl
 */
public final class ContextMenuUtils {

	/**
	 * Creates the menu populated with composable prototypes.
	 * 
	 * @param target
	 *            The {@link EObject} for which the menu shall be created.
	 * @param context
	 *            {@link IElementCompositionContext} to be used. This might be null.
	 * @param onlyForEditedObject
	 *            Determines whether there shall only be all MenuItems if the target is opened in as
	 *            (main) edited object in an editor. If 'true' is passed here, only IConnectors will
	 *            be added to elements that are not editedObject in the active editor.
	 * @return The menu items for composable elements.
	 */
	public static List<MenuItem> createPrototypeMenu(EObject target,
			IElementCompositionContext context, boolean onlyForEditedObject) {
		IPrototypeService pers = IPrototypeService.getInstance();
		IElementCompositorService ecs = IElementCompositorService.getInstance();
		IModelEditor<EObject> activeEditor =
				IModelEditorBindingService.getInstance().getActiveEditor();
		boolean isMenuForEditedObject =
				activeEditor != null && target == activeEditor.getEditedObject();

		List<MenuItem> result = new ArrayList<>();
		Menu newMenu = new Menu("New ...");
		Menu reuseMenu = new Menu("Reuse ...");

		List<Prototype> prototypes = pers.getComposablePrototypes(target.getClass());
		List<Prototype> prototypesForNewMenu = new ArrayList<>();
		for(Prototype prototype : prototypes) {
			EObject prototypeCopy = prototype.getPrototypeCopy();
			// If the context menu is created for the currently edited object, offer all composable
			// elements. Otherwise (i.e., if context menu is created for a structural element within
			// the currently edited object), offer only (composable) connectors.
			if((!onlyForEditedObject || isMenuForEditedObject ||
					prototypeCopy instanceof IConnector) &&
					ecs.canCompose(target, prototypeCopy, context)) {
				if(isCorrectReuseElement(prototypeCopy)) {
					MenuItem menuItem = createMenuItemForComposableElement(prototypeCopy, target,
							context, prototype.getName());
					String libraryName = getNameOfFirstRelatedLibrary(prototypeCopy);
					menuItem.setText("[" + libraryName + "] " + prototype.getName());
					reuseMenu.getItems().add(menuItem);
				} else {
					prototypesForNewMenu.add(prototype);
				}
			}
		}

		// Sort first by priority and then by name (lexicographical).
		prototypesForNewMenu.sort(new Comparator<Prototype>() {
			@Override
			public int compare(Prototype prototype1, Prototype prototype2) {
				int result = prototype1.getPriority() - prototype2.getPriority();
				if(result == 0) {
					// Negate integer to have the case that the order <A, B> corresponds to "A is
					// greater than B" (= positive)
					result = -prototype1.getName().compareTo(prototype2.getName());
				}
				return result;
			}
		});
		// Reverse to get a descending order (highest prio at first / on top).
		reverse(prototypesForNewMenu);

		for(Prototype prototype : prototypesForNewMenu) {
			EObject prototypeCopy = prototype.getPrototypeCopy();
			MenuItem menuItem = createMenuItemForComposableElement(prototypeCopy, target, context,
					prototype.getName());
			newMenu.getItems().add(menuItem);
		}

		if(!newMenu.getItems().isEmpty()) {
			result.add(newMenu);
		}
		if(!reuseMenu.getItems().isEmpty()) {
			result.add(reuseMenu);
		}

		return result;
	}

	/**
	 * Returns a menu item for the given element with which the element can be composed with the
	 * given target (in the given context).
	 * 
	 * @param composableElement
	 *            The element that should be composed.
	 * @param target
	 *            The target with which the element should be composed.
	 * @param context
	 *            The context of the composition.
	 * @param itemName
	 *            The name of the new menu item.
	 * @return The newly created menu item.
	 */
	private static MenuItem createMenuItemForComposableElement(EObject composableElement,
			EObject target, IElementCompositionContext context, String itemName) {
		ICommandStackService css = ICommandStackService.getInstance();
		IElementCompositorService ecs = IElementCompositorService.getInstance();
		MenuItem menuItem = new MenuItem(itemName);
		menuItem.setOnAction(evt -> {
			css.runAsCommand(target, () -> {
				ecs.compose(target, composableElement, context);
			});
		});
		return menuItem;
	}

	/**
	 * Creates the menu populated with composable prototypes.
	 * 
	 * Note:
	 * This method will only add all possible MenuItems if the target is opened as a (main)
	 * edited object in an editor. Otherwise only IConnectors will be added. If this is not the
	 * intended behavior, see also: createPrototypeMenu(EObject target, IElementCompositionContext
	 * context, boolean onlyForEditedObject)
	 * 
	 * @param target
	 *            The {@link EObject} for which the menu shall be created.
	 * @param context
	 *            {@link IElementCompositionContext} to be used. This might be null.
	 * @return The menu items for composable elements.
	 */
	public static List<MenuItem> createPrototypeMenu(EObject target,
			IElementCompositionContext context) {
		return createPrototypeMenu(target, context, true);
	}

	/** Creates a delete menu entry */
	public static MenuItem createDeleteMenu(EObject target) {
		IElementCompositorService ecs = IElementCompositorService.getInstance();
		ICommandStackService css = ICommandStackService.getInstance();

		MenuItem deleteMenuItem = new MenuItem("Delete");
		deleteMenuItem.setOnAction(evt -> {
			css.runAsCommand(target, () -> {
				ecs.decompose(target);
			});
		});

		return deleteMenuItem;
	}

	/** Creates an auto-layout menu entry for {@link IHierarchicElement}s. */
	public static MenuItem createAutoLayoutMenu(IHierarchicElement target) {
		final String MENU_NAME = "Automatic layout";
		MenuItem mi = new MenuItem(MENU_NAME);
		ICommandStackService css = ICommandStackService.getInstance();
		mi.setOnAction(evt -> {
			css.runAsCommand(target, () -> {
				IAutoLayouter layouter = new KielerAutoLayouter();
				layouter.performAutoLayout(target);
			});
		});
		return mi;
	}

	/** Creates the element composition context. */
	public static IElementCompositionContext createElementCompositionContext(EObject target,
			double x, double y, boolean isRoot, double zoom) {
		Point loc = createPoint((int)x, (int)y, "CompositionPoint");
		return new ElementDropContext(target, loc, isRoot, zoom);
	}
}
